#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: Jonas Schiessl

import nmpyc as mpc
from nmpyc.utils import mpc_convert

import numpy as np

from inspect import signature
import dill

class objective:
    """
    A class used to define the objective of the optimal control problem.
    
    The objective depends on the stage cost and optional on terminal cost and has the form

    .. math::

       J(t,x,u,N) := \sum_{k=0}^{N-1} \ell(t_k,x(t_k),u(t_k)) + F(t_N,x(t_N)).

    The values of the times :math:`t_k` are defined by initializing the :py:class:`nmpyc.system.system`.
    For the slightly different form of the objective in the discounted case see :py:attr:`~discount`.
    
    Parameters
    ----------
    stagecost : callable
        A function defining the stage cost of the optimal control problem. 
        Has to be of the form :math:`\ell(t,x,u)` or :math:`\ell(x,u)` in the :py:attr:`~autonomous` case.
        See also :py:attr:`stagecost`.
    terminalcost : callable, optional
        A function defining the terminal cost of the optimal control 
        problem. Has to be of the form :math:`F(t,x)` or :math:`F(x)` in the autonomous 
        case. If None, no terminal cost is added. 
        The default is None. See also :py:attr:`terminalcost`.
    
    """
    
    def __init__(self, stagecost, terminalcost=None):
        
        self.stagecost = stagecost
        self.terminalcost = terminalcost
        self._type = 'NLP'
        self._discont = 1.
        
    @property 
    def stagecost(self):
        """callable or list of array: Stage cost :math:`\ell(t,x,u)`.
        
        The return value of this attribute depends on how the 
        objective is initialized. 
        If it is initialized as a quadratic objective by :py:meth:`~LQP` a list 
        containing the arrays defining the stage cost are returned. 
        If the obejctive is initalized by possibly nonlinear callable functions 
        the function defining the stage cost is returned. 
        Note, that even if :py:attr:`~autonomous` is True the returned function depends 
        on the time :math:`t` and always has the form :math:`\ell(t,x,u)`.
        """
        return self._L
    
    @stagecost.setter 
    def stagecost(self, cost):
        if callable(cost):
            sig = signature(cost)
            params = sig.parameters
            if len(params) == 2:
                self._autonomous = True
                L = lambda t,x,u : cost(x,u)
            elif len(params) == 3:
                self._autonomous = False
                L = cost
            else:
                raise ValueError(
                    'stage cost must have two or three input arguments - not ' 
                    + str(len(params)))
        elif isinstance(cost, list):
            for i in range(3):
                if not isinstance(cost[i], mpc.array): 
                    raise TypeError(
                        'stage cost must be a list of arrays -  not ' 
                        + str(type(cost[i])))
            self._autonomous = True
            L = cost
        else: 
            raise TypeError(
                'stage cost must be callable or list of arrays - not ' 
                + str(type(cost)))
        self._L = L
        
    @property 
    def terminalcost(self):
        """callable or array: Terminal cost :math:`F(t,x)`.
        
        The return value of this attribute depends on how the 
        objective is initialized. 
        If it is initialized as a quadratic objective by :py:meth:`~LQP` the 
        array defining the terminal cost is returned. 
        If the obejctive is initalized by possibly nonlinear callable functions 
        the function defining the terminal cost is returned. 
        Note, that even if :py:attr:`~autonomous` is True the returned function depends 
        on the time :math:`t` and always has the form :math:`\ell(t,x,u)`.
        """
        return self._F
    
    @terminalcost.setter 
    def terminalcost(self, cost):
        F = None
        if callable(cost):
            sig = signature(cost)
            params = sig.parameters
            if len(params) == 1:
                F = lambda t,x : cost(x)
            elif len(params) == 2:
                self.autonomous = False
                F = cost
            else:
                raise ValueError(
                    'terminal cost must have two or three input arguments - not ' 
                    + str(len(params)))
        elif isinstance(cost,mpc.array):
            F = cost
        else: 
            if cost is not None:
                raise TypeError(
                    'terminal cost must be callable or arrays - not ' 
                    + str(type(cost)))
        self._F = F
        
    @property 
    def autonomous(self):
        """bool : If True, the objective is autonomous. 
        
        The stage cost and terminal cost 
        of the objective :math:`J(t,x,u,N)` 
        are not explicitly dependend on the time variable :math:`t`.
        In this case :math:`J(t,x,u,N)=J(x,u,N)` holds."""

        return self._autonomous
    
    @property 
    def type(self):
        """str : Indicating whether the objective is quadratic or nonlinear.
        
        If `LQP`, the objective is quadratic.
        Then the stage cost 
        is given by

        .. math::

           \ell(x,u) = x^TQx + u^TRu + 2x^TNx.

        and the terminal cost is given by

        .. math::

           F(x) = x^TPx.

        It also implies that the system is :py:attr:`~autonomous`.

        If the objective is not initialized as quadratic function with the :py:meth:`~LQP` 
        method this attribute holds the value `NLP`.
        """
        return self._type
    
    @property
    def discount(self):
        """float : The discount factor of the objective.
        
        For a discount factor :math:`\delta \in (0,1]` the 
        discounted objective function is given by

        .. math::
           
           J(t,x,u,N) = \sum_{k=0}^{N-1} \delta^k \ell(t_k,x(t_k),u(t_k)) + F(t_N,x(t_N)).

        By default the discount factor is equal to 1 :math:`\delta = 1`. Then, we name the problem undiscounted.
        The discount factor for the OCP of the MPC simulation can be set 
        when the :py:meth:`nmpyc.model.model.mpc` method is called.
        """
        return self._discont
    
    @discount.setter
    def discount(self,disc):
        if disc is None:
            disc = 1.
        if isinstance(disc, (int,float)):
            if disc <= 0 or disc > 1.: 
                raise ValueError('discount factor must be in (0,1]')
            self._discont = disc
        else: 
            raise TypeError(
                'discount factor must be of type integer or float - not ' 
                + str(type(disc)))  
        
    def __str__(self):
        
        string = ''
        
        string += 'autonomous: ' + str(self.autonomous) + '\n'
        
        terminal = False
        if self._F is not None: terminal = True
        string += 'terminalcost: ' + str(terminal) + '\n'
        
        lqp = False
        if self._type == 'LQP': lqp = True
        string += 'linear: ' + str(lqp) + '\n'
        
        string += 'discountfactor: ' + str(self.discount)
        
        return string
        
    @classmethod
    def LQP(cls, Q, R, N=None, P=None):
        """Initialize a quadratic objective.

        In this case the stage cost of the objective has the form

        .. math::
           \ell(x,u) = x^T Q x + u^T R u + 2 x^T Q u

        and the optional terminal cost is defined as 

        .. math::
           F(x,u) = x^T P x.

        In this case the objective is always :py:attr:`~autonomous`.

        Parameters
        ----------
        Q : array
            Matrix defining the cost of the state of the form :math:`x^TQx`.
        R : array
            Matrix defining the cost of the control of the form :math:`u^TRu`.
        N : array, optional
            Possible Matrix defining the mixed cost term of the form :math:`2x^TNu`. 
            The default is None.
        P : array, optional
            Posible Matrix defining the terminal cost of the form :math:`x^TPx`. 
            The default is None.

        Returns
        -------
        QP : objective
            nMPyC-objective class object suitable to define a linear 
            quadratic optimal control problem.

        """
        
        if N is None:
            N = mpc.zeros((Q.dim[0], R.dim[1]))
        if P is None:
            P = mpc.zeros((Q.dim[0], Q.dim[1]))
            
        if not isinstance(Q, mpc.array):
            raise TypeError('Q must be of type array - not ' + str(type(Q)))
        if not isinstance(R, mpc.array):
            raise TypeError('R must be of type array - not ' + str(type(R)))
        if isinstance(N, mpc.array):
            if N.dim != (Q.dim[0], R.dim[1]):
                raise ValueError(
                    'N has the flase dimension - ' 
                    + str(N.dim) + '!=' + str((Q.dim[0],R.dim[1])))
        else:
            raise TypeError('N must be of type array - not ' + str(type(N)))
        if isinstance(P, mpc.array):
            if P.dim != (Q.dim[0], Q.dim[1]):
                raise ValueError(
                    'N has the flase dimension - ' 
                    + str(P.dim) + '!=' + str((Q.dim[0], Q.dim[1])))
        else:
            raise TypeError('P must be of type array - not ' + str(type(P)))
        
        if Q.symbolic:
            raise ValueError(
                'A has to be purely numeric, but also has symbolic entries')   
        if R.symbolic:
            raise ValueError(
                'R has to be purely numeric, but also has symbolic entries')
        if N.symbolic:
            raise ValueError(
                'N has to be purely numeric, but also has symbolic entries')
        if P.symbolic:
            raise ValueError(
                'P has to be purely numeric, but also has symbolic entries')
        
        QP = cls([Q,R,N], P)
        QP._type = 'LQP'
        return QP
        
    def add_termianlcost(self, terminalcost):
        """Add terminal cost to the objective.
        
        The terminal cost must be a callable function of the form :math:`F(t,x)` 
        or :math:`F(x)` in the autonomous case. If terminal cost already exists they 
        will be over written.

        Parameters
        ----------
        terminalcost : callable
            A function defining the terminal cost of the optimal control 
            problem. Has to be of the form :math:`F(t,x)` or :math:`F(x)` in the autonomous 
            case.

        """
        
        self.terminalcost = terminalcost
        
    def J(self, t, x, u, N):
        """Evaluate objective function of the OCP.

        The objective function is assembled from the stage cost :math:`\ell(t,x,u)` 
        and optional terminal cost :math:`F(t,x)` and has the form 

        .. math::

           J(t,x,u,N) = \sum_{k=0}^{N-1} \delta^k \ell(t_k,x(t_k),u(t_k)) + F(t_N,x(t_N)).

        Where :math:`\delta \in (0,1]` is a possible discount factor, see :py:attr:`~discount`.

        Parameters
        ----------
        t : array
            Times instant at which the stage costs and terminal cost are evaluated.
        x : array
            State trajectory at which the stage cost and terminal cost are 
            evaluated.
        u : array
            Control sequence at which the stage cost is evaluated.
        N : int
            Maximum index up to which the stage cost are summed. 
            During the MPC iteration this index is equivalent to 
            the MPC horizon.

        Returns
        -------
        J : array
            Value of the objective function at the given input parameters.

        """
        
        if isinstance(N, int):
            if N <= 0 : 
                raise ValueError('horizon N must be greater than zero')
        else: 
            raise TypeError(
                'horizon N must be of type integer - not ' + str(type(N)))
        
        J = mpc.array([0])
        for k in range(N):
            J += self.stagecosts(t[k], x[:,k], u[:,k])*(self._discont**(k))
        J += self.endcosts(t[N],x[:,N])
        return J
    
    @mpc_convert
    def stagecosts(self,t,x,u):
        """Evaluate stage cost of the objective.

        Parameters
        ----------
        t : float
            Time instant at which the stage cost is evaluated.
        x : array
            Current state at which the stage cost is evaluated.
        u : array
            Current control at which the stage cost is evaluated.

        Returns
        -------
        array
            Stage cost evaluated at the given input values.
            
        """
            
        if self._type == 'NLP':
            return self._L(t,x,u)
        else:
            return (x.transpose()@self._L[0]@x 
                    + u.transpose()@self._L[1]@u 
                    + x.transpose()@self._L[2]@u*2)

    
    @mpc_convert
    def endcosts(self, t, x):
        """Evaluate termninal cost of the objective.

        Parameters
        ----------
        t : float
            Time instant at which the terminal cost is evaluated.
        x : array
            Current state at which the terminal cost is evaluated.

        Returns
        -------
        array
            Terminal cost evaluated at the given input values.

        """
        
        if self._F is None:
            return mpc.zeros(1)
        if self._type == 'NLP':
            return self._F(t, x)
        else:
            return x.transpose()@self._F@x
    
    def save(self, path):
        """Saving the objective to a given file with `dill <https://dill.readthedocs.io/en/latest/dill.html>`_.
        
        The path can be absolut or relative and 
        the ending of the file is arbitrary.

        Parameters
        ----------
        path : str
            String defining the path to the desired file. 



        For example

        >>> objective.save('objective.pickle')
        
        will create a file `objective.pickle` containing the nMPyC objective object.
        """

        
        with open(path, "wb") as output_file:
            dill.dump(self, output_file, -1)
    
    @classmethod
    def load(cls, path):
        """Loads a nMPyC objective object from a file.
        
        The specified path must lead to a file that was previously saved with
        :py:meth:`~save`.
        
        Parameters
        ----------
        path : str
            String defining the path to the file containing the nMPyC 
            objective object. 
            
            
        For example

        >>> objective.load('objective.pickle')
        
        will load the objective previously saved with :py:meth:`~save`.
            
        """

        
        try:
            with open(path, "rb") as input_file:
                e = dill.load(input_file)
        except:
            raise Exception(
                'Can not load objective from file. File not readable!')
            
        if not isinstance(e, objective):
            raise Exception(
                'Can not load objective from file. File does not contain a objective!')
            
        return e



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: Jonas Schiessl

import nmpyc as mpc 
from nmpyc.utils import mpc_convert

import numpy as np

from inspect import signature
import dill

class constraints:
    """
    Class used to define the constraints of the optimnal control problem.
    
    Support for nonlinear, linear and box constraints are implemented 
    and provided.

    To define the constraints, first, an empty object have to be created. 
    Then the individual constraints can be added with the help of the methods
    :py:meth:`add_bound` and :py:meth:`add_constr`.
    """
    
    def __init__(self):
        
        self._lower_bndx = None
        self._upper_bndx = None
        
        self._lower_bndu = None
        self._upper_bndu = None
        
        self._lower_bndend = None
        self._upper_bndend = None
        
        self._linear_constr = {'eq': [], 'ineq': [], 
                               'terminal_eq': [], 'terminal_ineq': []}
        self._nonlinear_constr = {'eq': [], 'ineq': [], 
                                  'terminal_eq': [], 'terminal_ineq': []}
        
        self._type = 'LQP'
        
    @property
    def lower_bndx(self):
        """array : Lower bound :math:`l_x \in \mathbb{R}^{nx}` of the state.
        
        For all states :math:`x(t_k)` the inequalities 

        .. math::

           x_i(t_k) \geq l_{x_i} \quad \\text{for } i = 1,\ldots,nx

        hold as a constraint.
        """
        return self._lower_bndx 
    
    @property
    def upper_bndx(self):
        """array : Upper bound :math:`u_x \in \mathbb{R}^{nx}` of the state.
        
        For all states :math:`x(t_k)` the inequalities 

        .. math::

           x_i(t_k) \leq u_{x_i} \quad \\text{for } i = 1,\ldots,nx

        hold as a constraint.
        """
        return self._upper_bndx 
    
    @property
    def lower_bndu(self):
        """array : Lower bound :math:`l_u \in \mathbb{R}^{nu}` for the control.
        
        For all controls :math:`u(t_k)` the inequalities 

        .. math::

           u_i(t_k) \geq l_{u_i} \quad \\text{for } i = 1,\ldots,nu

        hold as a constraint.
        """
        return self._lower_bndu
    
    @property
    def upper_bndu(self):
        """array : Upper bound :math:`u_u \in \mathbb{R}^{nu}` of the control.
        
        For all controls :math:`u(t_k)` the inequalities

        .. math::

           u_i(t_k) \leq u_{u_i} \quad \\text{for } i = 1,\ldots,nu

        hold as a constraint.
        """
        return self._upper_bndu 
    
    @property 
    def lower_bndend(self):
        """array : Lower bound :math:`l_x \in \mathbb{R}^{nx}` of the terminal state.
        
        For the terminal state :math:`x(t_N)` the inequality 

        .. math::

           x_i(t_N) \geq l_{x_i} \quad \\text{for } i = 1,\ldots,nx

        holds as a constraint.
        """
        return self._lower_bndend
    
    @property 
    def upper_bndend(self):
        """array : Upper bound :math:`u_x \in \mathbb{R}^{nx}` of the terminal state.
        
        For the terminal state :math:`x(t_N)` the inequalities

        .. math::

           x_i(t_N) \leq u_{x_i} \quad \\text{for } i = 1,\ldots,nx

        hold as a constraint.
        """
        return self._upper_bndend
    
    @property 
    def linear_constr(self):
        """dict : Collection of all linear constraints.
        
        This dictionary has the following form:

        >>> linear_constr = {'eq': [..], 'ineq': [..], 
        >>>                  'terminal_eq': [..], 'terminal_ineq': [..]}

        The arrays that define the constraints are saved as lists which are contained in the dictionary.
        For example

        >>> linear_constr['eq'][0]

        returns a list with the arrays :math:`H`, :math:`F` and :math:`h` defining the 
        first equality constraint

        .. math::

           Hx + Fu = h.
        """
        return self._linear_constr
    
    @property 
    def nonlinear_constr(self):
        """dict : Collection of all nonlinear constraints.
        
        This dictionary has the following form:

        >>> nonlinear_constr = {'eq': [..], 'ineq': [..], 
        >>>                     'terminal_eq': [..], 'terminal_ineq': [..]}

        In the lists contained in the dictionary the functions defining the 
        constraints are saved. 
        For example

        >>> nonlinear_constr['eq'][0]

        returns the function :math:`h(t,x,u)` defining the 
        first equality constraint

        .. math::

           h(t,x,u) = 0.
        """
        return self._nonlinear_constr
    
    @property 
    def type(self):
        """str : Indicating whether all constraints are linear.
        
        If `LQP`, all constraints are linear.
        Then all constraints are of the form

        .. math::

           Ex + Fu \leq h

        If at least one constraint is initialized as a nonlinear constraint 
        this attribute has the value `NLP`.
        """

        return self._type
        
    def __str__(self):
        string = ''
        
        nlbx = 0
        nubx = 0
        nlbe = 0
        nube = 0
        
        if self._lower_bndx is not None:
            for i in range(len(self._lower_bndx)):
                if self._lower_bndx[i] != -mpc.inf:
                    nlbx += 1
        if self._upper_bndx is not None:
            for i in range(len(self._upper_bndx)):
                if self.upper_bndx[i] != -mpc.inf:
                    nubx += 1
        if self._lower_bndend is not None:
            for i in range(len(self._lower_bndend)):
                if self.lower_bndend[i] != -mpc.inf:
                    nlbe += 1
        if self._upper_bndend is not None:
            for i in range(len(self._upper_bndend)):
                if self._upper_bndend[i] != -mpc.inf:
                    nube += 1
        
        string += 'number of lower bounds for state: ' + str(nlbx) +'\n'
        string += 'number of upper bounds for state: ' + str(nubx) +'\n'
        string += ('number of lower bounds for terminal state: ' 
                   + str(nlbe) 
                   +'\n')
        string += ('number of upper bounds for sterminal tate: ' 
                   + str(nube) 
                   +'\n')
        
        nlbu = 0
        nubu = 0
        
        if self._lower_bndu is not None:
            for i in range(len(self._lower_bndu)):
                if self._lower_bndu[i] != -mpc.inf:
                    nlbu += 1
        if self._upper_bndu is not None:
            for i in range(len(self._upper_bndu)):
                if self._upper_bndu[i] != -mpc.inf:
                    nubu += 1
        
        string += 'number of lower bounds for control: ' + str(nlbu) +'\n'
        string += 'number of upper bounds for control: ' + str(nubu) +'\n'
        
        for key in self._linear_constr.keys():
            string += ('number of linear ' 
                       + key 
                       + ' constraints: ' 
                       + str(len(self._linear_constr[key])) 
                       +'\n')
            
        for key in self._nonlinear_constr.keys():
            string += ('number of nonlinear ' 
                       + key 
                       + ' constraints: ' 
                       + str(len(self._nonlinear_constr[key])) 
                       +'\n')
        
        return string[:-1]
        
        
    def add_bound(self, bnd_type, variable, bound):
        """Add bounds as linear constraints to the OCP.

        Note while adding the bound it is not 
        checked if the bounds have the correct shape.
        This will be verified later during the optimization 
        progress.

        Parameters
        ----------
        bnd_type : str
            String defining whether the bound is a lower or upper bound.
        variable : str
            String defining on which variable the bound should be applied. 
            Possible values are *state*, *control* and *terminal*.
        bound : array
            Array containing the values of the bound.

        
        For example 

        >>> constraints.add_bound('lower', 'state', lbx)

        will add `lbx` as :py:attr:`~lower_bndx` while

        >>> constraints.add_bound('upper', 'terminal', ub_end)

        will add `ub_end` as :py:attr:`~upper_bndend`.

        """
        
        if not isinstance(bound, mpc.array):
            raise TypeError(
                'bound must be of type array - not ' + str(type(bound)))
        if bound.symbolic:
            raise ValueError(
                'bound has to be purely numeric,' 
                + ' but also has symbolic entries')
        if not bound.dim[1] == 1:
            if not bound.dim[0] == 1:
                print(bound.dim)
                raise ValueError(
                    'bound must be a flat array'
                    + ' but booth dimnensions are greter than 1')
            else:
                bound = bound.transpose()
            
        if not isinstance(bnd_type, str):
            raise TypeError(
                'bnd_type must be of type string - not ' 
                + str(type(bnd_type)))
        if not bnd_type in ['upper', 'lower']:
            raise ValueError(
                'bnd_type has to be upper or lower - not ' + bnd_type)
            
        if not isinstance(variable, str):
            raise TypeError(
                'variable must be of type string - not ' 
                + str(type(variable)))
        if not variable in ['state', 'control','terminal']:
            raise ValueError(
                'variable has to be state, control or terminal - not ' 
                + variable)
            
        if variable == 'state':
            if bnd_type == 'lower':
                self._lower_bndx = bound
            elif bnd_type == 'upper':
                self._upper_bndx = bound
        elif variable == 'control':
            if bnd_type == 'lower':
                self._lower_bndu = bound
            elif bnd_type == 'upper':
                self._upper_bndu = bound
        elif variable == 'terminal':
            if bnd_type == 'lower':
                self._lower_bndend = bound
            elif bnd_type == 'upper':
                self._upper_bndend = bound
                
    def add_constr(self, cons_type, *args):
        """Add linear or nonlinear constraints to the OCP.
        
        Nonlinear inequality constraints are of the form 

        .. math::

           g(t,x,u) \geq 0 \quad \\text{or} \quad g(x,u) \geq 0.

        Nonlinear equality constraints are of the form

        .. math::

           h(t,x,u) = 0 \quad \\text{or} \quad h(x,u) = 0.

        Linear inequality constraints are of the form 

        .. math::

           Ex + Fu \geq h.

        Linear equality constraints are of the form 

        .. math::

           Ex + Fu = h.

        For the form of terminal constrains see :py:meth:`~add_terminalconstr`.

        Parameters
        ----------
        cons_type : str
            String that defines the type of the constraints. 
            Possible values are *eq*, *ineq*, *terminal_eq* and *terminal_ineq*.
        *args : callable or arrays
            Function defining the (nonlinear) constraints or 
            arrays defining the linear constraints. 
            In the letter case the order of arguments are E, F, h and 
            if h is undefined this array is set to zero.



        For example 

        >>> constraints.add_constr('ineq', E, F, h)

        will add a linear inequality constraint to :py:attr:`linear_constr` while

        >>> constraints.add_constr('terminal_eq',h_end)

        will add a nonlinear equality terminal constraint to :py:attr:`nonlinear_constr`.

        """
        
        if cons_type  == 'terminal_eq':
            self.add_terminalconstr('eq', *args)
            return None
        elif cons_type  == 'terminal_ineq':
            self.add_terminalconstr('ineq', *args)
            return None
        
        if not isinstance(cons_type, str):
            raise TypeError()
        if cons_type not in ['eq', 'ineq']:
            raise ValueError()
        
        if len(args) == 1:
            linear = False
            if callable(args[0]):
                sig = signature(args[0])
                params = sig.parameters
                if len(params) == 2:
                    cons = lambda t,x,u : args[0](x, u)
                elif len(params) == 3:
                    cons = args[0]            
                else:
                    raise ValueError(
                        'callable non-terminal constraints must have' 
                        + ' two or three input arguments - not ' 
                        + str(len(params)))
            else:
                raise TypeError(
                    'constraints must be defined by a callable function ' 
                    + 'or as linear constraints by two arrays - not by ' 
                    + str(type(args[0])))
        elif len(args) == 2:
            linear = True
            for i in range(len(args)):
                if not isinstance(args[i], mpc.array): 
                    raise TypeError(
                        'linear constraints must be defined by arrays -  not ' 
                        + str(type(args[i])))
            h = np.zeros(args[0].dim[0])
            args = (args[0], args[1], h)
        elif len(args) == 3:
            linear = True
            for i in range(len(args)):
                if not isinstance(args[i], mpc.array): 
                    raise TypeError(
                        'linear constraints must be defined by arrays -  not ' 
                        + str(type(args[i])))
        else:
            raise ValueError(
                'given constraint must be a callable or two arrays')
        
        if linear:
            self._linear_constr[cons_type] += [args]
        else:
            self._nonlinear_constr[cons_type] += [mpc_convert(cons)]
            self._type = 'NLP'
            
    def add_terminalconstr(self, cons_type, *args):
        """Add linear or nonlinear terminal constraints to the OCP.
        
        Nonlinear terminal inequality constraints are of the form 

        .. math::

           g(t,x) \geq 0 \quad \\text{or} \quad g(x) \geq 0.

        Nonlinear terminal equality constraints are of the form

        .. math::

           h(t,x) = 0 \quad \\text{or} \quad h(x) = 0.

        Linear terminal inequality constraints are of the form 

        .. math::

           Ex \geq h.

        Linear terminal equality constraints are of the form 

        .. math::
        
           Ex = h.

        Parameters
        ----------
        cons_type : str
            String that defines the type of the terminal constraints. 
            Possible values are *eq* or *ineq*.
        *args : callable or arrays
            Function defining the (nonlinear) terminal constraints or 
            arrays defining the linear constraints. 
            In the letter case the order of arguments are E, h and 
            if h is undefined this array is set to zero.



        For example 

        >>> constraints.add_terminalconstr('ineq', E, F, h)

        will add a linear inequality terminal constraint to :py:attr:`linear_constr` while

        >>> constraints.add_constr('eq',h)

        will add a nonlinear equality terminal constraint to :py:attr:`nonlinear_constr`.
        """
        
        if not isinstance(cons_type, str):
            raise TypeError()
        if cons_type not in ['eq', 'ineq']: 
            raise ValueError()   
        if len(args) > 2:
            print(args)
            raise TypeError(
                'terminal constraints must be defined by a callable function ' 
                + 'or as linear constraints by an arrays')
        
        if len(args) == 1:
            if callable(args[0]):
                linear = False
                sig = signature(args[0])
                params = sig.parameters
                if len(params) == 1:
                    cons = lambda t,x : args[0](x)
                elif len(params) == 2:
                    cons = args[0]            
                else:
                    raise ValueError(
                        'callable terminal constraints must have' 
                        + ' one or two input arguments - not ' 
                        + str(len(params)))
            elif isinstance(args[0], mpc.array):
                linear = True
                h = np.zeros(args[0].dim[0])
                args = (args[0], h)
            else:
                raise ValueError(
                    'given constraint must be a callable or arrays')
        elif len(args) == 2:
            linear = True
            for i in range(len(args)):
                if not isinstance(args[i], mpc.array): 
                    raise TypeError(
                        'linear constraints must be defined by arrays -  not ' 
                        + str(type(args[i])))
        else:
            raise ValueError('given constraint must be a callable or arrays')
        
        cons_type = 'terminal_' + cons_type
        if linear:
            self._linear_constr[cons_type] += [args]
        else:
            self._nonlinear_constr[cons_type] += [mpc_convert(cons)]
            self._type = 'NLP'
                
    def save(self, path):
        """Saving the constraints to a given file with `dill <https://dill.readthedocs.io/en/latest/dill.html>`_.
        
        The path can be absolut or relative and 
        the ending of the file is arbitrary.

        Parameters
        ----------
        path : str
            String defining the path to the desired file. 



        For example

        >>> constraints.save('constraints.pickle')
        
        will create a file `constraints.pickle` containing the nMPyC constraints object.
        """
        
        with open(path, "wb") as output_file:
            dill.dump(self, output_file, -1)
    
    @classmethod
    def load(cls, path):
        """Loads a nMPyC constraints object from a file.
        
        The specified path must lead to a file that was previously saved with
        :py:meth:`~save`.
        
        Parameters
        ----------
        path : str
            String defining the path to the file containing the nMPyC 
            constraints object. 
            
            
        For example

        >>> constraints.load('constraints.pickle')
        
        will load the constraints previously saved with :py:meth:`~save`.
            
        """
        
        try:
            with open(path, "rb") as input_file:
                e = dill.load(input_file)
        except:
            raise Exception(
                'Can not load constraints from file. File not readable!')
            
        if not isinstance(e, constraints):
            raise Exception(
                'Can not load constraints from file. File does not cotain constraints!')
            
        return e
                
            



from setuptools import setup

setup(
    name='nmpyc',
    version='1.0.2',
    packages=['nmpyc'],
    author='Jonas Schießl, Lisa Krügel',
    author_email='nmpyc@uni-bayreuth.de',
    url='https://nmpyc.readthedocs.io/',
    description='Python library for solving optimal control problems via model predictive control',
    license='GNU Lesser General Public License version 3',
    long_description=open('README.md', 'r').read(),
    long_description_content_type="text/markdown",
    install_requires=[
        "casadi",
        "osqp",
        "numpy",
        "scipy",
        "matplotlib",
        "dill"
    ],
)



